/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*
 * $Log$
 */
#include <dev/spidigio.h>

/*!
 * \addtogroup xgSpiDigIo
 */
/*@{*/

/*!
 * \brief Loop for a approximately 25 microseconds.
 *
 * This call will not release the CPU and will not switch 
 * to another thread. However, because of absent thread 
 * switching, this delay time is very exact.
 *
 */
static inline void delay25us(void)
{
    u_char _cnt = 11;

    while(_cnt--)
        asm volatile("nop");
}

/*!
 * \brief Toggle SPI clock to perform a shift.
 */
static inline void ShiftDigital(void)
{
    sbi(PORTD, 7);
    asm volatile("nop\n\tnop\n\tnop\n\tnop\n");
    cbi(PORTD, 7);
}

/*!
 * \brief Return the number of shifts required to get an expected bit 
 * value at the input.
 *
 * When calling this function, it is assumed, that the shift 
 * register is already filled with the complement of the input bit.
 *
 * If the search mode is zero, then the function will return
 * the number of shifts until the bit value at the input appears 
 * at the output. This can be used to detect the total size of the
 * shift register.
 *
 * If the search mode is set to one, then the input strobe will be 
 * toggled before shifting and the function returns the number of 
 * shifts until the first unmodified bit appears.  This can be used 
 * to detect the number of inputs.
 *
 * \param num   Total number of shifts, should be 8, 16, 24 or 32.
 * \param bit   Input bit, either 0 or 1.
 * \param smoed Search mode.
 *
 * \return The number of shifts.
 */
static u_char CountDigitalShifts(u_char num, u_char bit, u_char smode)
{
    u_char i;
    u_char rc = 0;

    /*
     * Toggle input strobe if we are searching the first
     * unmodified bit.
     */
    if(smode) {
        cbi(PORTB, 7);
        delay25us();
        sbi(PORTB, 7);
    }

    /*
     * Set the input bit.
     */
    if(bit)
        sbi(PORTD, 5);
    else
        cbi(PORTD, 5);
    delay25us();

    /*
     * Do the requested number of shifts and watch
     * the requested bit position.
     */
    for(i = 0; i < num; i++) {
        delay25us();
        if(bit_is_set(PIND, 6)) {
            if(bit) {
                if(smode) 
                    rc = i + 1;
                else if(rc == 0) 
                    rc = i + 1;
            }
        }
        else {
            if(bit == 0) {
                if(smode) 
                    rc = i + 1;
                else if(rc == 0)
                    rc = i + 1;
            }
        }            
        ShiftDigital();
    }
    return rc;
}

/*!
 * \brief Initialize the digital I/O interface.
 *
 * Automatically detects the number of digital inputs and outputs.
 *
 * \param inputs  Pointer to an 8-bit value, where the number of 
 *                detected inputs will be stored.
 * \param outputs Pointer to an 8-bit value, where the number of 
 *                detected outputs will be stored.
 */
void SpiDigitalInit(u_char *inputs, u_char *outputs)
{
    u_char total = 0;
    u_char i;
    u_char cnt;

    /*
     * Set serial data output 
     * line direction.
     */
    sbi(DDRD, 5);

    /*
     * Set serial data input 
     * line direction and 
     * enable pullup.
     */
    sbi(PORTD, 6);
    cbi(DDRD, 6);

    /*
     * Clock.
     */
    sbi(DDRD, 7);

    /*
     * Strobe.
     */
    sbi(DDRB, 5);

    sbi(PORTB, 7);
    sbi(DDRB, 7);

    /*
     * Fill the shift register with zeros.
     * Then shift in ones until the first
     * appears on the output. This gives 
     * us the total size plus one of the 
     * shift register.
     */
    CountDigitalShifts(32, 0, 0);
    total = CountDigitalShifts(32, 1, 0) - 1;

    /*
     * Determine the number of inputs. We do 
     * this five times for zeros and ones and 
     * take the maximum count. This way we 
     * compensate changing inputs while 
     * counting.
     */
    *inputs = 0;
    for(i = 0; i < 5; i++) {
        if((cnt = CountDigitalShifts(total, 0, 1)) > *inputs)
            *inputs = cnt;
        if((cnt = CountDigitalShifts(total, 1, 1)) > *inputs)
            *inputs = cnt;
    }
    *outputs = total - *inputs;
}

/*!
 * \brief Query digital inputs and outputs.
 *
 * If both inputs and outputs are installed, typically the
 * inputs are represented by the most signigficant bits.
 *
 * SpiDigitalInit() must have been called by the application
 * before calling this function.
 *
 * \param num Number of bits to query, either 8, 16, 24 or 32.
 *
 * \return Binary value of the requested inputs.
 */
u_long SpiDigitalGet(u_char num)
{
    u_char i;
    u_long bits = 0;
    
    /*
     * Toggle input strobe.
     */
    cbi(PORTB, 7);
    delay25us();
    sbi(PORTB, 7);
    
    for(i = 0; i < num; i++) {
        delay25us();
        bits <<= 1;
        if(bit_is_set(PIND, 6))
            bits |= 1;
        ShiftDigital();
    }
    return bits;
}

/*!
 * \brief Set digital outputs.
 *
 * Although possible, it makes no sense to set digital inputs.
 *
 * SpiDigitalInit() must have been called by the application
 * before calling this function.
 *
 * \param num Number of bits to set, either 8, 16, 24 or 32.
 * \param bits The bit value to be set
 */
void SpiDigitalSet(u_char num, u_long bits)
{
    u_char i;
    u_long mask;

    if(num) {
        mask = 1UL << (num - 1);
        for(i = 0; i < num; i++) {
            
            /*
             * Set output bit.
             */
            if(bits & mask)
                sbi(PORTD, 5);
            else
                cbi(PORTD, 5);
            delay25us();
                
            ShiftDigital();

            mask >>= 1;
        }
        
        /*
         * Toggle strobe.
         */
        sbi(PORTB, 5);
        delay25us();
        cbi(PORTB, 5);
    }
}

/*@}*/
