#include <stdio.h>
#include <windows.h>

#include "getopt.h"
#include "hexfile.h"
#include "ispport.h"
#include "pdriver.h"

typedef struct {
    char *name;
    u_short sign;
    u_long  pags;
    u_long  prgs;
    u_long  eeps;
} PARTINFO;

PARTINFO known_parts[] = {
  { "AT90S1200", 0x9001,   0,   1024,   64 },
  { "ATtiny12",  0x9005,   0,   1024,   64 },
  { "ATtiny15",  0x9006,   0,   1024,   64 },
  { "AT90S2313", 0x9101,   0,   2048,  128 },
  { "AT90S2343", 0x9103,   0,   2048,  128 },
  { "AT90S2323", 0x9102,   0,   2048,  128 },
  { "AT90S2333", 0x9105,   0,   2048,  128 },
  { "ATtiny22",  0x9106,   0,   2048,  128 },
  { "AT90S4433", 0x9203,   0,   4096,  256 },
  { "AT90S4414", 0x9201,   0,   4096,  256 },
  { "AT90S4434", 0x9202,   0,   4096,  256 },
  { "AT90S8515", 0x9301,   0,   8192,  512 },
  { "AT90S8535", 0x9303,   0,   8192,  512 },
  { "ATmega8",   0x9307,  64,   8192,  512 },
  { "ATmega161", 0x9401, 128,  16384,  512 },
  { "ATmega163", 0x9402, 128,  16384,  512 },
  { "ATmega16",  0x9403, 128,  16384,  512 },
  { "ATmega323", 0x9501, 128,  32768, 1024 },
  { "ATmega64",  0x9602, 256,  65536, 2048 },
  { "ATmega103", 0x9701, 256, 131072, 4096 },
  { "ATmega128", 0x9702, 256, 131072, 4096 },
  { "ATmega103", 0x0101, 256, 131072, 4096 },
  { 0,           0xFFFF, 256, 131072, 4096 },
};

static u_char progMem[131072];
int verbose = 0;
int erase = 0;

/*!
 * \brief Send an ISP command.
 */
void IspCmd(u_char cmd_rsp[4])
{
    int i;
    u_int m;
    u_int r;

    for(i = 0; i < 4; i++) {
        r = cmd_rsp[i];
        if(verbose > 1)
            fprintf(stderr, "%02X", r);
        for(m = 0x80; m; m >>= 1) 
            if(IspPortData(r & m))
                r |= m;
            else
                r &= ~m;
        if(verbose > 1)
            fprintf(stderr, "-%02X ", r);
        cmd_rsp[i] = r;
    }
    if(verbose > 1)
        fprintf(stderr, "\n");
}

/*
 * Detect part.
 */
PARTINFO *AvrDetectPart(void)
{
    u_short sign;
    u_char cmd_rsp[4];
    int i;

    /*
     * Read signature high byte at address 1.
     */
    cmd_rsp[0] = 0x30;
    cmd_rsp[2] = 1;
    IspCmd(cmd_rsp);
    sign = cmd_rsp[3] << 8;

    /*
     * Read signature low byte at address 2.
     */
    cmd_rsp[0] = 0x30;
    cmd_rsp[2] = 2;
    IspCmd(cmd_rsp);
    sign += cmd_rsp[3];

    for(i = 0; known_parts[i].name; i++) 
        if(known_parts[i].sign == sign) 
            return &known_parts[i];
    known_parts[i].sign = sign;
    return &known_parts[i];
}

/*!
 * \brief Enable AVR programming mode.
 */
int AvrProgEnable(void)
{
    int i;

    for(i = 0; i < 32; i++) {
        u_char prg[4] = { 0xAC, 0x53, 0xFF, 0xFF };

        IspCmd(prg);
        if(prg[2] == 0x53) 
            return 0;
        IspPortClock(1);
        Sleep(1);
        IspPortClock(0);
        Sleep(1);
    }
    return -1;
}

/*!
 * \brief Read program memory.
 */
void AvrReadProgMem(u_char *mem)
{
    int h;
    int l;
    u_char cmd_rsp[4];
    u_long addr = 0;

    for(h = 0; h < 256; h++) {
        for(l = 0; l < 256; l++) {

            /*
             * Read program memory low byte.
             */
            cmd_rsp[0] = 0x20;
            cmd_rsp[1] = h;
            cmd_rsp[2] = l;
            IspCmd(cmd_rsp);
            mem[addr++] = cmd_rsp[3];

            /*
             * Read program memory high byte.
             */
            cmd_rsp[0] = 0x28;
            cmd_rsp[1] = h;
            cmd_rsp[2] = l;
            IspCmd(cmd_rsp);
            mem[addr++] = cmd_rsp[3];
        }
    }
}

u_long AvrReadFuses(void)
{
    u_long fuses;
    u_char cmd_rsp[4];

    /*
     * Read extended fuses.
     */
    cmd_rsp[0] = 0x50;
    cmd_rsp[1] = 0x08;
    IspCmd(cmd_rsp);
    fuses = cmd_rsp[3] << 16;

    /*
     * Read fuses high byte.
     */
    cmd_rsp[0] = 0x58;
    cmd_rsp[1] = 0x08;
    IspCmd(cmd_rsp);
    fuses |= cmd_rsp[3] << 8;

    /*
     * Read fuses low byte.
     */
    cmd_rsp[0] = 0x50;
    cmd_rsp[1] = 0;
    IspCmd(cmd_rsp);
    fuses |= cmd_rsp[3];

    return fuses;
}

/*!
 * \brief Send chip erase command.
 */
void AvrChipErase(void)
{
    u_char cmd_rsp[4] = { 0xAC, 0x80, 0x00, 0x00 };

    if(verbose)
        fprintf(stderr, "Chip erased...");
    IspCmd(cmd_rsp);
    Sleep(1);
    if(verbose)
        fprintf(stderr, "done\n");
}

/*!
 * \brief Write a program memory page.
 */
void AvrWriteProgPage(u_int pg, const u_char *pp)
{
    u_char cmd[4];
    u_int i;
    
    for(i = 0; i < 128; i++) {
        /*
         * Load page memory low byte.
         */
        cmd[0] = 0x40;
        cmd[2] = i;
        cmd[3] = *pp++;
        IspCmd(cmd);

        /*
         * Load page memory high byte.
         */
        cmd[0] = 0x48;
        cmd[2] = i;
        cmd[3] = *pp++;
        IspCmd(cmd);
    }

    /*
     * Write program memory page.
     */
    cmd[0] = 0x4C;
    cmd[1] = pg >> 1;
    cmd[2] = (pg & 1) << 7;
    IspCmd(cmd);
    Sleep(50);
}

int AvrVerifyProgPage(u_int pg, const u_char *pp)
{
    int i;
    u_char cmd_rsp[4];
    u_long addr = pg * 128;

    for(i = 0; i < 128; i++) {

        /*
         * Read program memory low byte.
         */
        cmd_rsp[0] = 0x20;
        cmd_rsp[1] = (u_char)(addr >> 8);
        cmd_rsp[2] = (u_char)addr;
        IspCmd(cmd_rsp);
        if(*pp != cmd_rsp[3]) {
            fprintf(stderr, "Verify failed at addr %04X: %02X %02X\n", addr, *pp, cmd_rsp[3]);
            return -1;
        }
        pp++;

        /*
         * Read program memory high byte.
         */
        cmd_rsp[0] = 0x28;
        cmd_rsp[1] = (u_char)(addr >> 8);
        cmd_rsp[2] = (u_char)addr;
        IspCmd(cmd_rsp);
        if(*pp != cmd_rsp[3]) {
            fprintf(stderr, "Verify failed at addr %04X: %02X %02X\n", addr, *pp, cmd_rsp[3]);
            return -1;
        }
        pp++;
        addr++;
    }
    return 0;
}

int AvrWriteProgMem(u_char *mem)
{
    int rc = 0;
    u_int pg;
    u_int i;
    u_long addr;


    /*
     * Loop for all 512 pages.
     */
    if(verbose)
        fprintf(stderr, "Programming program memory...");
    addr = 0;
    for(pg = 0; pg < 512 && rc == 0; pg++) {
        /*
         * Program this page if any byte is not 0xFF.
         */
        for(i = 0; i < 256; i++) {
            if(mem[addr + i] != 0xFF) {
                AvrWriteProgPage(pg, &mem[addr]);
                if(AvrVerifyProgPage(pg, &mem[addr]))
                    rc = -1;
                break;
            }
        }
        addr += 256;
    }
    if(verbose)
        fprintf(stderr, "done\n");
    return rc;
}

/*!
 * \brief Toggle reset line.
 */
void AvrReset(void)
{
    IspPortReset(1);
    Sleep(1);
    IspPortReset(0);
    Sleep(20);
}

/*
 * \brief Set reset line high.
 */
void AvrRun(void)
{
    IspPortReset(1);
    IspPortEnable(0);
}

void PrintFuseBits(u_long fuses)
{
    fprintf(stderr, "Fuse bits: ");

    /*
     * Extended.
     */
    if((fuses & 0x020000) == 0)
        fprintf(stderr, "M103C ");
    if((fuses & 0x010000) == 0)
        fprintf(stderr, "WDTON ");

    /*
     * High.
     */
    if((fuses & 0x8000) == 0)
        fprintf(stderr, "OCDEN ");
    if((fuses & 0x4000) == 0)
        fprintf(stderr, "JTAGEN ");
    if((fuses & 0x2000) == 0)
        fprintf(stderr, "SPIEN ");
    if((fuses & 0x1000) == 0)
        fprintf(stderr, "CKOPT ");
    if((fuses & 0x0800) == 0)
        fprintf(stderr, "EESAVE ");
    if((fuses & 0x0400) == 0)
        fprintf(stderr, "BOOTSZ1 ");
    if((fuses & 0x0200) == 0)
        fprintf(stderr, "BOOTSZ0 ");
    if((fuses & 0x0100) == 0)
        fprintf(stderr, "BOOTRST ");

    /*
     * Low.
     */
    if((fuses & 0x80) == 0)
        fprintf(stderr, "BODLEVEL ");
    if((fuses & 0x40) == 0)
        fprintf(stderr, "BODEN ");
    if((fuses & 0x20) == 0)
        fprintf(stderr, "SUT1 ");
    if((fuses & 0x10) == 0)
        fprintf(stderr, "SUT0 ");
    if((fuses & 0x08) == 0)
        fprintf(stderr, "CKSEL3 ");
    if((fuses & 0x04) == 0)
        fprintf(stderr, "CKSEL2 ");
    if((fuses & 0x02) == 0)
        fprintf(stderr, "CKSEL1 ");
    if((fuses & 0x01) == 0)
        fprintf(stderr, "CKSEL0 ");

    fprintf(stderr, "\n");
}

void usage(void)
{
    fputs("Usage: nutisp OPTIONS FILENAME\n"
          "OPTIONS:\n"
          "-e      erase chip\n"
          "-i PATH install port driver\n"
          "-r      remove port driver\n"
          "-v      verbose\n"
        , stderr);
}

int main(int argc, char **argv)
{
    int option;
    int i;
    int rc = 0;
    PARTINFO *part;
#if 0
    u_long fusebits;
#endif

    while((option = getopt(argc, argv, "ei:rv?")) != EOF) {
        switch(option){
        case 'e':
            erase++;
            break;
        case 'i':
            if(PortDriverInstall(optarg))
                fprintf(stderr, "Installation failed\n");
            return 1;
        case 'r':
            if(PortDriverRemove())
                fprintf(stderr, "De-installation failed\n");
            return 1;
        case 'v':
            verbose++;
            break;
        default:
            usage();
            return 1;
        }
    }

    /*
     * Collect remaining parameters as hex files.
     */
    argc -= optind;
    argv += optind;
    memset(progMem, 0xFF, sizeof(progMem));
    for(i = 0; i < argc; i++)
        ReadIntel(argv[0], progMem);


    if(IspPortInit())
        return 3;

    IspPortEnable(1);
    IspPortClock(0);

    IspPortReset(0);
    AvrReset();

    if(AvrProgEnable()) {
        fprintf(stderr, "No AVR chip detected\n");
        AvrRun();
        return 4;
    }
    part = AvrDetectPart();
    fprintf(stderr, "Detected %s\n", part->name ? part->name : "unknown chip");

    if(erase) {
        AvrChipErase();
        AvrReset();
        if(AvrProgEnable()) {
            fprintf(stderr, "No AVR chip detected\n");
            AvrRun();
            return 4;
        }
    }

#if 0
    fusebits = AvrReadFuses();
    PrintFuseBits(fusebits);
#endif

    if(AvrWriteProgMem(progMem)) {
        fprintf(stderr, "Programming failed\n");
        rc = 2;
    }
    AvrRun();

    return 0;
}
